import logging
from functools import cached_property
from typing import Any

from pydantic import (
    computed_field,
    AnyHttpUrl,
    EmailStr,
    ConfigDict,
)
from authlib.integrations.starlette_client.apps import StarletteOAuth2App
from sqlmodel import SQLModel, Field

logger = logging.getLogger("oidc-test")


class Role(SQLModel, extra="ignore"):
    name: str


class UserBase(SQLModel, extra="ignore"):
    id: str | None = None
    sid: str | None = None
    name: str | None = None
    email: EmailStr | None = None
    picture: AnyHttpUrl | None = None
    roles: list[Role] = []


class User(UserBase):
    model_config = ConfigDict(arbitrary_types_allowed=True)  # type:ignore

    sub: str = Field(
        description="""subject id of the user given by the oidc provider,
                            also the key for the database 'table'""",
    )
    userinfo: dict = {}
    access_token: str | None = None
    access_token_decoded: dict[str, Any] | None = None
    oidc_provider: StarletteOAuth2App | None = None

    @computed_field
    @cached_property
    def roles_as_set(self) -> set[str]:
        return set([role.name for role in self.roles])

    def has_scope(self, scope: str) -> bool:
        """Check if the scope is present in user info or access token"""
        info_scopes = self.userinfo.get("scope", "").split(" ")
        try:
            access_token_scopes = self.decode_access_token().get("scope", "").split(" ")
        except Exception as err:
            logger.info(f"Access token cannot be decoded: {err}")
            access_token_scopes = []
        return scope in set(info_scopes + access_token_scopes)

    def decode_access_token(self):
        assert self.access_token is not None
        assert self.oidc_provider is not None
        assert self.oidc_provider.name is not None
        from .settings import oidc_providers_settings

        return oidc_providers_settings[self.oidc_provider.name].decode(self.access_token)
